import 'dart:async';

/// Token for cancelling a pending operation.
///
/// - create your token (final token = CancellationToken())
/// - attach to a pending task future (myTask = token.attach(myTask))
/// - cancel in any moment (token.cancel('just changed my mind'))
class CancellationToken {
  Completer? _completer;

  /// Check if there is a pending operation attached. When this is false, [cancel] will do nothing
  bool get active => _completer != null;

  /// Create a cancellation token
  CancellationToken();

  /// Create a cancellation token that will automatically cancel after given [duration] from when [attach] is called
  factory CancellationToken.timeout(Duration duration) => _TimeOutCancellationToken(duration);

  /// Attach [token] to [task] if not null
  static Future<T> wrap<T>(Future<T> task, CancellationToken? token) => token == null ? task : token.attach(task);

  /// Wrap [task] with this token.
  /// After this call you should use the returned future instead of [task]
  /// The returned future will raise a [CancelException] when cancellation is requested
  Future<T> attach<T>(Future<T> task) {
    assert(_completer == null, 'already attached to another future');
    final completer = _completer = Completer<T>();
    task.then(completer.complete, onError: completer.completeError);
    return completer.future..whenComplete(() => _completer = null);
  }

  /// Cancel pending operation
  void cancel([String message = 'operation canceled']) => _completer?.completeError(CancelException._(message));
}

/// Exception raised when a cancellation is requested
class CancelException implements Exception {
  final String message;

  const CancelException._(this.message);

  @override
  String toString() => message;
}

class _TimeOutCancellationToken extends CancellationToken {
  final Duration duration;

  _TimeOutCancellationToken(this.duration);

  @override
  Future<T> attach<T>(Future<T> future) =>
      super.attach(future)..whenComplete(Timer(duration, () => cancel('operation canceled after $duration')).cancel);
}
