import 'dart:async';

import 'package:flutter/foundation.dart';

/// Mutual exclusion.
/// Acquire and release lock
class Mutex {
  Completer<void>? _completer;

  bool get running => _completer != null;

  Future<T> lock<T>(Future<T> Function() function) => acquire().then((_) async {
        try {
          return await function();
        } finally {
          release();
        }
      });

  Future<void> acquire() {
    final c = _completer;
    if (c == null) {
      _completer = Completer();
      return SynchronousFuture(null);
    } else {
      return c.future.then((_) => acquire());
    }
  }

  void release() {
    final c = _completer;
    if (c == null) throw StateError('cannot release, nothing acquired');
    c.complete();
    _completer = null;
  }
}

/// Run async function successfully exactly once
class RunOnce extends Mutex {
  bool _done = false;

  Future<bool> call(Future Function() function) => _done || running
      ? SynchronousFuture(false)
      : lock<bool>(() async {
          if (_done) return false;
          await function();
          return _done = true;
        });
}

/// Run async function, every second call made to this function will be ignored until the first call completes
class RunFirst extends Mutex {
  Future<bool> call(Future Function() function) =>
      running ? SynchronousFuture(false) : lock(function).then((_) => true);
}

/// Run async function after [delay], if called twice in that delay only the last call will be scheduled,
/// previous calls will be ignored
class RunDelayed extends Mutex {
  final Duration delay;
  late DateTime _lastCall;

  RunDelayed(this.delay);

  Future<bool> call(Future Function() function) {
    final now = _lastCall = DateTime.now();
    return Future.delayed(
        delay,
        () => now != _lastCall
            ? SynchronousFuture(false)
            : lock<bool>(() async {
                if (now != _lastCall) return false;
                await function();
                return true;
              }));
  }
}
